#include "Raytracer.h"

#include <vector>
#include <algorithm>

#include "src/shader/Shader.h"

#include "src/scene/SurfaceList.h"

Raytracer::Raytracer(int& renderResolution, int& previewResolution, int& maxRayDepth)
    : renderResolution(renderResolution)
    , previewResolution(previewResolution)
    , maxRayDepth(maxRayDepth)
{
    // Precalculates the rotation matrices for the direction vector because doing this in the
    // distribution ray tracing function will get too slow:
    for(int x = -1; x <= 1; ++x){
        for(int y = -1; y <= 1; ++y){
            distributedRotationMatrices.push_back(Mat4f::rotationX(0.05f * x) * Mat4f::rotationY(0.05f * y));
        }
    }

    // Create raytracing thread which calls the render function if wanted:
    raytraceThread = std::thread([this](){
        while(!waitForTerminate){
            if(!displayPreviewImage){
                isRendering = false;
                raytraceSemaphore.acquire();
                isRendering = true;
            }

            bool renderInPreviewMode = previewMode;

            resizeImagesIfNecessary();

            if(waitForTerminate)
                break;

            render(renderInPreviewMode);
        }
    });

    raytraceSemaphore.release();
}

Raytracer::~Raytracer(){
    waitForTerminate = true;

    raytraceSemaphore.release();
    raytraceThread.join();
}

float Raytracer::getProgress(){
    return progress;
}

void Raytracer::initPixelIDMap(std::vector<unsigned int>& pixelIDMap, unsigned int size){
    pixelIDMap = std::vector<unsigned int>(size);
    for(unsigned int i = 0; i < size; ++i){
        pixelIDMap[i] = i;
    }

    std::random_shuffle(pixelIDMap.begin(), pixelIDMap.end());
}

void Raytracer::resizeImagesIfNecessary(){
    int renderSize = renderResolution * renderResolution;
    if(int(renderedImage.size()) != renderSize){
        initPixelIDMap(renderPixelIDMap, renderSize);
        renderedImage = std::vector<Vec4f>(renderSize);
    }

    int previewSize = previewResolution * previewResolution;
    if(int(previewImage.size()) != previewSize){
        initPixelIDMap(previewPixelIDMap, previewSize);
        previewImage = std::vector<Vec4f>(previewSize);
    }
}

void Raytracer::copyPreviewToFull(){
    #pragma omp parallel num_threads(16)
    {
        float scaleFactor = float(previewResolution) / renderResolution;

        #pragma omp for
        for(int x = 0; x < int(renderResolution); ++x){
            for(int y = 0; y < int(renderResolution); ++y){
                int pX = int(x * scaleFactor);
                int pY = int(y * scaleFactor);

                renderedImage[y * renderResolution + x] = previewImage[pY * previewResolution + pX];
            }
        }
    }
}


void Raytracer::render(bool renderInPreviewMode){
    // Copy local variables:
    int resolution = renderResolution;
    std::vector<Vec4f>* image = &renderedImage;
    std::vector<unsigned int>* pixelIDMap = &renderPixelIDMap;

    shouldInterrupt = false;
    // If preview mode, we have to use other variables:
    if(renderInPreviewMode){
        resolution = previewResolution;
        image = &previewImage;
        pixelIDMap = &previewPixelIDMap;
        progress = 0;
    } else {
        copyPreviewToFull();
        displayPreviewImage = false;
    }

    // Generate a ray for each pixel and raytrace:
    for(int x = 0; x < resolution; ++x){

        #pragma omp parallel num_threads(16)
        {
            #pragma omp for
            for(int y = 0; y < resolution; ++y){
                // Pixel ID:
                unsigned int id = pixelIDMap->at(y * resolution + x);
                unsigned int rX = id % resolution;
                unsigned int rY = id / resolution;

                // The ray which should be traced for this pixel:
                Ray ray = scene.camera.generateRay(rX, rY, resolution, resolution);

                // Store the color into the image array:
                Vec4f pixColor = traceRay(ray, maxRayDepth);

                pixColor.w = 1.f;
                image->at(id) = pixColor;
            }

        }

        if(!renderInPreviewMode && previewMode){
            break;
        }

        if(shouldInterrupt){
            shouldInterrupt = false;
            break;
        }

        // Update progress bar:
        if(!renderInPreviewMode)
            progress = x / float(resolution);
    }

    // Avoid double rendering in full resolution:
    raytraceSemaphore.setCount(0);
}


void Raytracer::setPreviewMode(bool mode) {
    previewMode = mode;

    // If preview mode is activated, ensure that rendering gets reactivated:
    if(mode){
        if(raytraceSemaphore.getCount() <= 0)
            raytraceSemaphore.release();

        displayPreviewImage = true;
    }
}

void Raytracer::setScene(Scene pScene){
    scene = pScene;

    // Trigger repainting:
    setPreviewMode(true);
}

Vec4f Raytracer::traceRay(Ray ray, int depth){
    // If maximum depth is reached, just return black:
    if (depth == 0)
        return Vec4f();

    // Perform an intersection test agaist the surface (all surfaces):
    Intersection intersection = scene.surface.intersect(ray, 0.001f, 20.f);

    // If there is an intersection, shoot secondary rays and shade:
    if (intersection.hasIntersection) {
        Vec4f color = shade(intersection, ray);

        if(intersection.shader->hasReflective()){
            color += intersection.shader->getReflectiveColor() * traceRay(ray.reflected(intersection.hitPoint, intersection.normal), depth - 1);
        }

        if(intersection.shader->hasRefractive()){
            Ray refractedRay = ray.refracted(intersection.hitPoint, intersection.normal, intersection.shader->getRefractiveIndex());
            if(refractedRay.isValid()){
                color += intersection.shader->getRefractiveColor() * traceRay(refractedRay, depth - 1);
            }
        }

        return color;
    }

    // Otherwise, return background color:
    return backgroundColor;
}

Vec4f Raytracer::shade(Intersection& intersection, Ray& ray){
    //return intersection.normal;

    Vec4f color = intersection.shader->getAmbientColor();

    // Iterate over all lights:
    for(PointLight& light : scene.lights){
        if(!scene.surface.intersect(Ray(intersection.hitPoint, light.getPosition() - intersection.hitPoint), 0.01272f, 0.99856f).hasIntersection)
        color += intersection.shader->shade(light, intersection, ray);
    }

    return color;
}
