// © CGVR 2023, Andre Muehlenbrock (muehlenb@uni-bremen.de)

#include "CSGPrimitives.h"

#include <algorithm>

/**
 * Performs an intersection test against the CSGSphere.
 */

std::vector<CSGInterval> CSGSphere::intersectPrimitive(Ray& ray) const
{
    // QVector3D center = transformation * QVector3D(0.0, 0.0, 0.0);
    float radius = 0.5;

    // Transform the ray direction and it's origin in model space:
    Vec4f originMS = inverseTransform * ray.origin;
    Vec4f dirMS = inverseTransform * ray.direction;

    Vec4f v3Diff = originMS;
    v3Diff.w = 0;

    // The coefficients of the quadratic equation
    float A = dirMS.dot(dirMS);
    float B = 2 * dirMS.dot(v3Diff);
    float C = v3Diff.dot(v3Diff) - radius * radius;

    float discriminant = B*B-4*A*C;

    if ( discriminant <= FLT_EPSILON )
        return std::vector<CSGInterval>();

    // check the two solutions
    float sqrtdis = sqrt( discriminant );
    float t0 = (-B - sqrtdis ) / ( 2*A );
    float t1 = (-B + sqrtdis ) / ( 2*A );

    std::vector<CSGInterval> result;
    CSGInterval r;

    Vec4f t0Vec = (originMS + dirMS * t0);
    t0Vec.w = 0;
    r.aT = t0;
    r.aHitPoint = ray.evaluate(t0);
    r.aNormal = (normalMatrix * t0Vec).normalized();

    Vec4f t1Vec = (originMS + dirMS * t1);
    t1Vec.w = 0;
    r.bT = t1;
    r.bHitPoint = ray.evaluate(t1);
    r.bNormal = (normalMatrix * t1Vec).normalized();

    result.push_back(r);
    return result;
}

/**
 * Calculates the normal for the CSGCube in model space.
 *
 * @param dir   The direction from the cube center to the hit point.
 */

Vec4f CSGCube::calculateNormalMS(Vec4f dir) const {

    Vec4f result;
    if(std::abs(dir.x) > std::abs(dir.y) && std::abs(dir.x) > std::abs(dir.z))
        result = Vec4f(1.f,0.f,0.f,0.f) * (dir.x < 0 ? -1.f : 1.f);
    else if(std::abs(dir.y) > std::abs(dir.z))
        result = Vec4f(0.f,1.f,0.f,0.f) * (dir.y < 0 ? -1.f : 1.f);
    else
        result = Vec4f(0.f,0.f,1.f,0.f) * (dir.z < 0 ? -1.f : 1.f);

    return result;
}

/**
 * Performs an intersection test against the CSGCube.
 */

std::vector<CSGInterval> CSGCube::intersectPrimitive(Ray&ray) const
{
    Vec4f originMS = inverseTransform * ray.origin;
    Vec4f dirMS = inverseTransform * ray.direction;

    Vec4f min = Vec4f(-0.5f, -0.5f, -0.5f);
    Vec4f max = Vec4f(0.5f, 0.5f, 0.5f);

    float t1 = (min.x - originMS.x) / dirMS.x;
    float t2 = (max.x - originMS.x) / dirMS.x;
    float t3 = (min.y - originMS.y) / dirMS.y;
    float t4 = (max.y - originMS.y) / dirMS.y;
    float t5 = (min.z - originMS.z) / dirMS.z;
    float t6 = (max.z - originMS.z) / dirMS.z;

    float tMin = std::max(std::max(std::min(t1, t2), std::min(t3, t4)), std::min(t5, t6));
    float tMax = std::min(std::min(std::max(t1, t2), std::max(t3, t4)), std::max(t5, t6));

    if(tMin > tMax)
        return std::vector<CSGInterval>();

    std::vector<CSGInterval> result;

    CSGInterval interval;
    interval.aT = tMin;
    interval.aHitPoint = ray.evaluate(interval.aT);
    interval.aNormal =  (normalMatrix * calculateNormalMS(originMS + dirMS * interval.aT)).normalized();

    interval.bT = tMax;
    interval.bHitPoint = ray.evaluate(interval.bT);
    interval.bNormal = (normalMatrix * calculateNormalMS(originMS + dirMS * interval.bT)).normalized();

    result.push_back(interval);

    return result;
}
